"""
Base classes for osc-tiny specific exceptions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""


class OscError(Exception):
    """
    Base class for expcetions to be raised by ``osctiny``
    """
